import datetime as dt
import os
from pathlib import Path


from .cfg import Config
from .pipeline import Pipeline, PipelineManager
from .scheduler import SchedulerManager
import rich


def init(
    name: str | None = None,
    base_dir: str | None = None,
    conf_path: str = "conf",
    pipelines_path: str = "pipelines",
):
    if name is None:
        name = Path.cwd().name
        base_dir = Path.cwd().parent

    if base_dir is None:
        base_dir = Path.cwd()

    os.makedirs(os.path.join(base_dir, name, "conf/pipelines"), exist_ok=True)
    os.makedirs(os.path.join(base_dir, name, "pipelines"), exist_ok=True)

    cfg = Config.load(base_dir=os.path.join(base_dir, name))

    with open(os.path.join(base_dir, name, "README.md"), "w") as f:
        f.write(
            f"# {name.replace('_', ' ').upper()}\n\n"
            f"**created with FlowerPower**\n\n*{dt.datetime.now().strftime('%Y-%m-%d %H:%M:%S')}*\n\n"
        )
    cfg.save()
    os.chdir(os.path.join(base_dir, name))

    rich.print(
        f"\n✨ Initialized FlowerPower project [bold blue]{name}[/bold blue] at [italic green]{base_dir}[/italic green]\n"
    )

    rich.print("""[yellow]Getting Started:[/yellow]

    📦 It is recommended to use the project manager [bold cyan]`uv`[/bold cyan] to manage the dependenvies of your project.

    🔧 Install uv:
        [dim]Run:[/dim] [bold white]pip install uv[/bold white]
        [dim]More options:[/dim] [blue underline]https://docs.astral.sh/uv/getting-started/installation/[/blue underline]

    🚀 Initialize your project:
        [dim]Run the following in your project directory:[/dim]
        [bold white]uv init --app --no-readme --vcs git[/bold white]
    """)


def find_pipelines(cls):
    """Find all pipeline modules in the project's pipelines directory."""
    pipeline_path = Path("pipelines")
    if not pipeline_path.exists():
        return []

    pipelines = []
    for file in pipeline_path.glob("*.py"):
        if file.name.startswith("_"):
            continue

        module_name = file.stem
        try:
            pipeline = Pipeline(module_name)
            pipelines.append(pipeline)
        except Exception as e:
            rich.print(f"[red]Error loading pipeline {module_name}: {str(e)}[/red]")

    return pipelines


Pipeline.find_pipelines = classmethod(find_pipelines)


def list_pipelines():
    pipelines = Pipeline.find_pipelines()
    if not pipelines:
        rich.print("\n📭 [yellow]No pipelines found in this project[/yellow]\n")
        return

    rich.print("\n🌸 [bold magenta]Available Pipelines:[/bold magenta]\n")
    table = rich.table.Table(show_header=True, header_style="bold cyan")
    table.add_column("Name")
    table.add_column("Description")
    table.add_column("Status")

    for pipeline in pipelines:
        status = (
            "[green]Active[/green]" if pipeline.is_active() else "[red]Inactive[/red]"
        )
        table.add_row(
            pipeline.name, pipeline.description or "[dim]No description[/dim]", status
        )

    rich.print(table)
    rich.print()



import datetime as dt
import os
from dataclasses import asdict, dataclass, field

import yaml
from hamilton.function_modifiers import source, value
from munch import Munch, munchify, unmunchify


@dataclass
class BaseConfig:
    def to_dict(self):
        return unmunchify(asdict(self))  # .__dict__)

    def to_yaml(self, path: str):
        with open(path, "w") as f:
            yaml.dump(self.to_dict(), f, default_flow_style=False)

    @classmethod
    def from_dict(cls, d: dict | Munch):
        return cls(**d)

    @classmethod
    def from_yaml(cls, path: str):
        with open(path) as f:
            return cls.from_dict(yaml.full_load(f))


@dataclass
class PipelineRunConfig(BaseConfig):
    final_vars: list[str] = field(default_factory=list)
    inputs: dict | Munch = field(default_factory=dict)
    executor: str | None = None
    with_tracker: bool = False

    def __post_init__(self):
        if isinstance(self.inputs, dict):
            self.inputs = munchify(self.inputs)


@dataclass
class PipelineScheduleTriggerConfig(BaseConfig):
    type: str | None = None
    crontab: str | None = None
    year: str | int | None = None
    month: str | int | None = None
    weeks: int | float = 0
    week: str | int | None = None
    days: int | float = 0
    day: str | int | None = None
    minutes: int | float = 0
    minute: str | int | None = None
    seconds: int | float = 0
    second: str | int | None = None
    start_time: dt.datetime | None = None
    end_time: dt.datetime | None = None
    timezone: str | None = None


@dataclass
class PipelineScheduleRunConfig(BaseConfig):
    id: str | None = None
    executor: str | None = None
    paused: bool = False
    coalesce: str = "latest"  # other options are "all" and "earliest"
    misfire_grace_time: int | float | dt.timedelta | None = None
    max_jitter: int | float | dt.timedelta | None = None
    max_running_jobs: int | None = None
    conflict_poilcy: str | None = (
        "do_nothing"  # other options are "replace" and "exception"
    )


@dataclass
class PipelineScheduleConfig(BaseConfig):
    run: PipelineScheduleRunConfig = field(default_factory=PipelineScheduleRunConfig)
    trigger: PipelineScheduleTriggerConfig = field(
        default_factory=PipelineScheduleTriggerConfig
    )


@dataclass
class PipelineTrackerConfig(BaseConfig):
    project_id: int | None = None
    version: str | None = None
    dag_name: str | None = None
    tags: dict | Munch = field(default_factory=dict)

    def __post_init__(self):
        if isinstance(self.tags, dict):
            self.tags = munchify(self.tags)


@dataclass
class PipelineConfig(BaseConfig):
    name: str | None = None
    run: PipelineRunConfig = field(default_factory=PipelineRunConfig)
    schedule: PipelineScheduleConfig = field(default_factory=PipelineScheduleConfig)
    func: dict | Munch = field(default_factory=dict)
    tracker: PipelineTrackerConfig = field(default_factory=PipelineTrackerConfig)

    def __post_init__(self):
        self.run = PipelineRunConfig(
            **self.run if isinstance(self.run, dict | Munch) else self.run.to_dict()
        )
        self.schedule = PipelineScheduleConfig(
            **self.schedule
            if isinstance(self.schedule, dict | Munch)
            else self.schedule.to_dict()
        )
        self.tracker = PipelineTrackerConfig(
            **self.tracker
            if isinstance(self.tracker, dict | Munch)
            else self.tracker.to_dict()
        )
        if isinstance(self.func, dict):
            self.hamilton_func_params = munchify(self.f_args_to_ht_params(self.func))
            self.func = munchify(self.func)

    def to_dict(self):
        d = asdict(self)
        # d.pop("name")
        return unmunchify(d)

    def to_yaml(self, path: str):
        with open(path, "w") as f:
            d = self.to_dict()
            d.pop("name")
            yaml.dump(d, f, default_flow_style=False)

    @classmethod
    def from_dict(cls, name: str, d: dict | Munch):
        d.update({"name": name})
        return cls(**d)

    @classmethod
    def from_yaml(cls, name: str, path: str):
        with open(path) as f:
            return cls.from_dict(name, yaml.full_load(f))

    @staticmethod
    def f_args_to_ht_params(d: dict) -> dict:
        """Coverts a dictionary of function arguments to Hamilton function parameters"""

        def transform_recursive(val, original_dict):
            # If it's a dictionary, recursively transform its values
            if isinstance(val, dict):
                return {
                    k: transform_recursive(v, original_dict) for k, v in val.items()
                }
            # If it's a string and matches a key in the original dictionary
            elif isinstance(val, str) and val in original_dict:
                return source(val)
            # For all other values
            else:
                return value(val)

        # result = {}
        # for key, val in d.items():
        #     result[key] = {
        #         key: val
        #     }

        # Step 1: Replace each value with a dictionary containing key and value
        result = {k: {k: d[k]} for k in d}

        # Step 2 & 3: Transform all values recursively
        return {k: transform_recursive(v, d) for k, v in result.items()}


@dataclass
class ProjectSchedulerConfig(BaseConfig):
    data_store: dict | Munch = field(default_factory=dict)
    event_broker: dict | Munch = field(default_factory=dict)
    cleanup_interval: int | float | dt.timedelta = 900  # int in secods
    max_concurrent_jobs: int = 100

    def __post_init__(self):
        if isinstance(self.data_store, dict):
            self.data_store = munchify(self.data_store)
        if isinstance(self.event_broker, dict):
            self.event_broker = munchify(self.event_broker)


@dataclass
class ProjectTrackerConfig(BaseConfig):
    username: str | None = None
    api_url: str = "http://localhost:8241"
    ui_url: str = "http://localhost:8242"
    api_key: str | None = None


@dataclass
class ProjectConfig(BaseConfig):
    scheduler: ProjectSchedulerConfig = field(default_factory=ProjectSchedulerConfig)
    tracker: ProjectTrackerConfig = field(default_factory=ProjectTrackerConfig)

    def __post_init__(self):
        self.scheduler = ProjectSchedulerConfig(
            **self.scheduler
            if isinstance(self.scheduler, dict | Munch)
            else self.scheduler.to_dict()
        )
        self.tracker = ProjectTrackerConfig(
            **self.tracker
            if isinstance(self.tracker, dict | Munch)
            else self.tracker.to_dict()
        )


@dataclass
class Config(BaseConfig):
    base_dir: str | None = None
    pipeline: PipelineConfig = field(default_factory=PipelineConfig)
    project: ProjectConfig = field(default_factory=ProjectConfig)

    def __post_init__(self):
        self.pipeline = PipelineConfig(
            **self.pipeline
            if isinstance(self.pipeline, dict | Munch)
            else self.pipeline.to_dict()
        )
        self.project = ProjectConfig(
            **self.project
            if isinstance(self.project, dict | Munch)
            else self.project.to_dict()
        )

    @classmethod
    def load(cls, base_dir: str, pipeline_name: str | None = None):
        if os.path.exists(os.path.join(base_dir, "conf/project.yml")):
            project = ProjectConfig.from_yaml(
                path=os.path.join(base_dir, "conf/project.yml")
            )
        else:
            project = ProjectConfig()

        if pipeline_name is not None:
            if os.path.exists(
                os.path.join(base_dir, "conf", f"pipelines/{pipeline_name}.yml")
            ):
                pipeline = PipelineConfig.from_yaml(
                    name=pipeline_name,
                    path=os.path.join(
                        base_dir, "conf", f"pipelines/{pipeline_name}.yml"
                    ),
                )
            else:
                pipeline = PipelineConfig(name=pipeline_name)
        else:
            pipeline = PipelineConfig(name=pipeline_name)

        return cls(base_dir=base_dir, pipeline=pipeline, project=project)

    def save(self):
        if not os.path.exists(os.path.join(self.base_dir, "conf")):
            os.makedirs(os.path.join(self.base_dir, "conf"))

        if self.pipeline.name is not None:
            self.pipeline.to_yaml(
                os.path.join(self.base_dir, f"conf/pipelines/{self.pipeline.name}.yml")
            )
        self.project.to_yaml(os.path.join(self.base_dir, "conf/project.yml"))


def load(base_dir: str, pipeline_name: str | None = None):
    return Config.load(base_dir, pipeline_name)


def save(config: Config):
    config.save()
